@@ -1,7 +1,8 @@
+import random
+
 import cv2
 import numpy as np
 from PIL import Image, ImageDraw
-import random
 
 from .color_utils import map_colors
 from .transform import Transform
@@ -13,6 +14,7 @@ class LidarCamera:
     def __init__(self, camera_id):
         self.id = camera_id
         self.pose = Transform(self.world2cam)
+        self.world_poses = None
         self.K = np.eye(3, dtype=np.float32)
         self.D = np.zeros(6, dtype=np.float32)
         self.model = 'brown_conrady'
@@ -110,7 +112,7 @@ class LidarCamera:
         if K is not None:
             self.K = np.array(K[:3, :3])
         if D is not None:
-            self.D = np.pad(D, (0, 6 - len(D)), 'constant')
+            self.D = D
         if model is not None:
             self.model = model
         if scale_factor is not None:
@@ -127,6 +129,8 @@ class LidarCamera:
             self.K[0, 2] = kwargs['cx']
         if 'cy' in kwargs:
             self.K[1, 2] = kwargs['cy']
+        if 'world_poses' in kwargs:
+            self.world_poses = kwargs['world_poses']
 
     def apply_transform(self, transform: Transform):
         self.pose = transform @ self.pose
