@@ -8,19 +8,18 @@ from io import BytesIO
 from typing import Any, MutableMapping, List, Dict
 
 import boto3
+import matplotlib.pyplot as plt
 import numpy as np
 import pandas as pd
 import ujson
-from pyquaternion import Quaternion
 from PIL import Image, ImageEnhance
-from scaleapi import Task
-import matplotlib.pyplot as plt
 from matplotlib.widgets import Slider, Button
-import matplotlib.gridspec as gridspec
+from pyquaternion import Quaternion
+from scaleapi import Task
 
 from .camera import LidarCamera
-from .transform import Transform
 from .connectors import Importer
+from .transform import Transform
 
 UPLOAD_POOL_SIZE = 8
 s3 = boto3.client('s3')
@@ -81,6 +80,7 @@ class LidarImage:
         self.file = None
         self.transform = None
         self.metadata = None
+        self.timestamp = None
 
     def load_file(self, file):
         if not isinstance(file, str):
@@ -139,6 +139,8 @@ class LidarFrame:
         self.cameras: pd.Series[Any, LidarCamera] = cameras
         self.images: pd.Series[Any, LidarImage] = pd.Series(dtype=object)
         self.points: np.ndarray = np.zeros((0, 5), dtype=float)
+        self.radar_points: np.ndarray = np.zeros((0, 3), dtype=float)
+        self.colors: np.ndarray = np.zeros((0, 3), dtype=float)
         self.transform = Transform()
 
     def get_image(self, camera_id) -> LidarImage:
@@ -157,6 +159,10 @@ class LidarFrame:
     def add_points_from_connector(self, connector: Importer, transform: Transform = None, intensity=1, sensor_id=0):
         self.add_points(connector.data, transform, intensity, sensor_id)
 
+    def add_radar_points(self, points: np.array):
+        assert np.array(points).shape[1] == 3, 'Radar points length is not 3'
+        self.radar_points = points
+
     def add_points(self, points: np.array, transform: Transform = None, intensity=1, sensor_id=0):
         if points.ndim == 1:
             points = np.array([points])
@@ -169,6 +175,9 @@ class LidarFrame:
 
         self.points = np.vstack([self.points, points])
 
+    def add_colors(self, colors: np.ndarray):
+        self.colors = np.vstack([self.colors, colors])
+
     def add_debug_lines(self, intensity=1, length=5, device=0):
         x_line = np.array([np.array([length * k / 100, 0, 0]) for k in range(0, 100)])
         for camera in self.cameras:
@@ -230,6 +239,7 @@ class LidarFrame:
         bApply = Button(axApply, 'Apply changes')
         bApply.on_clicked(update)
         plt.show()
+
     def get_filename(self):
         return "frame-%s.json" % self.id
 
@@ -246,7 +256,12 @@ class LidarFrame:
 
         def format_image(camera):
             image = self.images[camera.id]
-            wct = (image.transform or self.transform) @ camera.pose
+            if camera.world_poses:
+                wct = camera.world_poses[self.id]
+            else:
+                wct = (image.transform or self.transform) @ camera.pose
+
+            D = camera.D
             result = dict(
                 position=format_point(wct.position),
                 heading=format_quaternion(wct.quaternion),
@@ -257,31 +272,47 @@ class LidarFrame:
                 cx=camera.cx,
                 cy=camera.cy,
                 skew=camera.skew,
-                k1=float(camera.D[0]),
-                k2=float(camera.D[1]),
-                p1=float(camera.D[2]),
-                p2=float(camera.D[3]),
-                k3=float(camera.D[4]),
-                k4=float(camera.D[5]),
+                k1=float(D[0]),
+                k2=float(D[1]),
+                p1=float(D[2]),
+                p2=float(D[3]),
+                k3=float(D[4]) if len(D) == 5 else 0,
+                k4=float(D[5]) if len(D) == 6 else 0,
+                k5=float(D[6]) if len(D) == 7 else 0,
+                k6=float(D[7]) if len(D) == 8 else 0,
                 scale_factor=camera.scale_factor
             )
             if image.metadata:
                 result['metadata'] = image.metadata
+            if image.timestamp:
+                result['timestamp'] = image.timestamp
             return result
 
         images_json = self.cameras[self.images.index].apply(format_image).to_json(orient='records')
         points_json = pd.DataFrame(self.get_world_points(), columns=['x', 'y', 'z', 'i', 'd']) \
             .to_json(double_precision=4, orient='records', date_format=None)
 
-        out = ujson.dumps({
+        colors_json = pd.Series(self.colors.reshape(-1).astype(np.uint32)).to_json(orient='values', date_format=None)
+
+        radar_points_json = ujson.dumps(list(np.array([{'position': format_point(row[0]), 'direction': format_point(row[1]), 'size': row[2][0]} for row in self.radar_points])))
+        frame_object = {
             'images': "__IMAGES__",
             'points': "__POINTS__",
             'device_position': format_point(self.transform.position),
             'device_heading': format_quaternion(self.transform.quaternion)
-        })
+        }
 
+        if len(self.colors) > 0:
+            frame_object['point_colors'] = "__COLORS__"
+
+        if len(self.radar_points) > 0:
+            frame_object['radar_points'] = "__RADAR_POINTS__"
+
+        out = ujson.dumps(frame_object)
         out = out.replace('"__IMAGES__"', images_json)
         out = out.replace('"__POINTS__"', points_json)
+        out = out.replace('"__COLORS__"', colors_json)
+        out = out.replace('"__RADAR_POINTS__"', radar_points_json)
 
         return out
 
@@ -486,4 +517,4 @@ class LidarScene:
             from .task import LidarSegmentation
             return LidarSegmentation.from_scene(self, template)
         else:
-            assert NotImplemented(f'Specified task_type {task_type} is not supported')
+            raise NotImplemented(f'Specified task_type {task_type} is not supported')
